// Copyright (C) 2017 James Jonathan Cook
// This program is distributed under the terms of the GNU Lesser General Public License.
//
// This program was written by James Jonathan Cook (jmsck55@gmail.com)
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU Lesser General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

/*
Optimized for pthread GCC

LINUX:
callback: #define euthread_function(NAME, ARG) void * NAME (void * ARG)
new_mutex: #define new_mutex(NAME) pthread_mutex_t NAME = PTHREAD_MUTEX_INITIALIZER
new_condition: #define new_condition(NAME) pthread_cond_t NAME = PTHREAD_COND_INITIALIZER;

createThread: int pthread_create(pthread_t *thread, const pthread_attr_t *attr, void *(*start_routine) (void *), void *arg);
joinThread: int pthread_join(pthread_t thread, void **retval);
yieldThread: int sched_yield(void);
exitThread: void pthread_exit(void *retval);
selfThread: pthread_t pthread_self(void);

initSection: #define initSection(foo)
freeSection: #define freeSection(foo)
lockSection: pthread_mutex_lock( &mutex1 );
unlockSection: pthread_mutex_unlock( &mutex1 );

initCondition: #define initCondition(foo)
sleepCondition: int pthread_cond_wait(pthread_cond_t *restrict cond, pthread_mutex_t *restrict mutex);
wakeCondition: int pthread_cond_signal(pthread_cond_t *cond);


*/


#include "euthreadpthread.h"

void foo(void) {
	puts("euthreadpthread.c\n");
}


int createThread(pthread_t * thread, void *(*start_routine) (void *), void * arg) {
	//void *(*start_routine) (void *)
	//int pthread_create(pthread_t *thread, const pthread_attr_t *attr, void *(*start_routine) (void *), void *arg);
	
	int iret;
	iret = pthread_create(thread, NULL, start_routine, arg);
	if (iret != 0) return (-iret); // return negative on error.
	return *(int *)thread;
}

int joinThread(int thread, void * retval) {
	int iret;
	iret = pthread_join((pthread_t) thread, (void **)retval);
	
	return iret;
}

int yieldThread(void) {
	//pthread_yield();
	return sched_yield();
}

void exitThread(void * retval) {
	pthread_exit(retval);
}

int selfThread(void) {
	return (int)pthread_self();
}

//#define newSection(foo)

//#define freeSection(foo)

void lockSection(pthread_mutex_t * p) {
	pthread_mutex_lock( p );
}

void unlockSection(pthread_mutex_t * p) {
	pthread_mutex_unlock( p );
}

//#define newCondition(foo)

int sleepCondition(pthread_cond_t *restrict cond, pthread_mutex_t *restrict mutex) {
	pthread_cond_wait( cond, mutex );
	return 1;
}

int wakeCondition(pthread_cond_t *cond) {
	return pthread_cond_signal( cond );
}

