// Copyright (C) 2017 James Jonathan Cook
// This program is distributed under the terms of the GNU Lesser General Public License.
//
// This program was written by James Jonathan Cook (jmsck55@gmail.com)
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU Lesser General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

/*
Optimized for MS Windows

WIN32:
pthread_t: typedef DWORD pthread_t;
callback: #define euthread_function(NAME, ARG) DWORD WINAPI NAME( LPVOID ARG )
new_mutex: #define new_mutex(NAME) CRITICAL_SECTION NAME
new_condition: #define new_condition(NAME) CONDITION_VARIABLE NAME

createThread: HANDLE WINAPI CreateThread (LPSECURITY_ATTRIBUTES lpThreadAttributes,
 SIZE_T dwStackSize, LPTHREAD_START_ROUTINE lpStartAddress, LPVOID lpParameter,
 DWORD dwCreationFlags, LPDWORD lpThreadId);
joinThread: DWORD WINAPI WaitForSingleObject(HANDLE hHandle, DWORD dwMilliseconds);
	BOOL WINAPI GetExitCodeThread(HANDLE hThread, LPDWORD lpExitCode);
yieldThread: BOOL WINAPI SwitchToThread(void);
exitThread: VOID WINAPI ExitThread(DWORD dwExitCode);
selfThread: DWORD WINAPI GetCurrentThreadId(void);

initSection: void InitializeCriticalSection(LPCRITICAL_SECTION lpCriticalSection);
freeSection: void DeleteCriticalSection(&CriticalSection);
lockSection: void EnterCriticalSection(&CriticalSection);
unlockSection: void LeaveCriticalSection(&CriticalSection);

initCondition: void InitializeConditionVariable(&condition_var);
sleepCondition: int SleepConditionVariableCS(&condition_var, &CriticalSection, INFINITE);
wakeCondition: int WakeConditionVariable(&condition_var);

*/

#include "euthreadwin.h"

void foo(void) {
	puts("euthreadwin.c\n");
}

void init(void) {

kernel32 = LoadLibrary("kernel32.dll");

xCloseHandle = (FuncCloseHandle)GetProcAddress(kernel32, "CloseHandle");
xCreateThread = (FuncCreateThread)GetProcAddress(kernel32, "CreateThread");
xWaitForSingleObject = (FuncWaitForSingleObject)GetProcAddress(kernel32, "WaitForSingleObject");
xGetExitCodeThread = (FuncGetExitCodeThread)GetProcAddress(kernel32, "GetExitCodeThread");
xSwitchToThread = (FuncSwitchToThread)GetProcAddress(kernel32, "SwitchToThread");
xExitThread = (FuncExitThread)GetProcAddress(kernel32, "ExitThread");
xGetCurrentThreadId = (FuncGetCurrentThreadId)GetProcAddress(kernel32, "GetCurrentThreadId");
xInitializeCriticalSection = (FuncInitializeCriticalSection)GetProcAddress(kernel32, "InitializeCriticalSection");
xDeleteCriticalSection = (FuncDeleteCriticalSection)GetProcAddress(kernel32, "DeleteCriticalSection");
xEnterCriticalSection = (FuncEnterCriticalSection)GetProcAddress(kernel32, "EnterCriticalSection");
xLeaveCriticalSection = (FuncLeaveCriticalSection)GetProcAddress(kernel32, "LeaveCriticalSection");
xInitializeConditionVariable = (FuncInitializeConditionVariable)GetProcAddress(kernel32, "InitializeConditionVariable");
xSleepConditionVariableCS = (FuncSleepConditionVariableCS)GetProcAddress(kernel32, "SleepConditionVariableCS");
xWakeConditionVariable = (FuncWakeConditionVariable)GetProcAddress(kernel32, "WakeConditionVariable");

}

int closeHandle(HANDLE h) {
	return (int)(xCloseHandle)(h);
}

int createThread(pthread_t * pThreadId, void * ThreadProc, void * arg) {
	HANDLE hThread;
	hThread = (xCreateThread)(NULL, 1, (LPTHREAD_START_ROUTINE)ThreadProc,
	 (LPVOID)arg, STACK_SIZE_PARAM_IS_A_RESERVATION, (LPDWORD)pThreadId);
	if (hThread == NULL) return (-1); // return negative on error.
	return (int)hThread;
}

int joinThread(int hThread, void * lpExitCode) {
	DWORD iret;
	iret = (xWaitForSingleObject)((HANDLE)hThread, INFINITE);
	if (lpExitCode != NULL) (xGetExitCodeThread)((HANDLE)hThread, (LPDWORD)lpExitCode);
	
	return (int)iret;
}

int yieldThread(void) {
	return (int)(xSwitchToThread)();
}

void exitThread(void * dwExitCode) {
	// You still need to close the handle of the thread.
	//(xCloseHandle)(hThread1);
	(xExitThread)((DWORD) dwExitCode);
}

int selfThread(void) {
	return (int)(xGetCurrentThreadId)();
}

void initSection(LPCRITICAL_SECTION lpCriticalSection) {
	(xInitializeCriticalSection)(lpCriticalSection);
}
void freeSection(LPCRITICAL_SECTION lpCriticalSection) {
	(xDeleteCriticalSection)(lpCriticalSection);
}

void lockSection(LPCRITICAL_SECTION pCriticalSection) {
	(xEnterCriticalSection)(pCriticalSection);
}

void unlockSection(LPCRITICAL_SECTION pCriticalSection) {
	(xLeaveCriticalSection)(pCriticalSection);
}

void initCondition(PCONDITION_VARIABLE pConditionVariable) {
	(xInitializeConditionVariable)(pConditionVariable);
}
int sleepCondition(PCONDITION_VARIABLE pConditionVariable, LPCRITICAL_SECTION pCriticalSection) {
	return (int)(xSleepConditionVariableCS)(pConditionVariable, pCriticalSection, INFINITE);
}
int wakeCondition(PCONDITION_VARIABLE pConditionVariable) {
	(xWakeConditionVariable)(pConditionVariable);
	return 0;
}

