-- Copyright (C) 2017 James Jonathan Cook
-- This program is distributed under the terms of the GNU Lesser General Public License.
--
-- This program was written by James Jonathan Cook (jmsck55@gmail.com)
--
--    This program is free software: you can redistribute it and/or modify
--    it under the terms of the GNU Lesser General Public License as published by
--    the Free Software Foundation, either version 3 of the License, or
--    (at your option) any later version.
--
--    This program is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--    GNU Lesser General Public License for more details.
--
--    You should have received a copy of the GNU Lesser General Public License
--    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--

namespace euthreadwin

include std/dll.e
include std/machine.e

public constant SIZE_OF_CRITICAL_SECTION = 24
public constant SIZE_OF_CONDITION_VARIABLE = 4
public constant STACK_SIZE_PARAM_IS_A_RESERVATION = 65536
public constant INFINITE = -1

public atom kernel32
public integer
	xCloseHandle,
	xCreateThread,
	xWaitForSingleObject,
	xGetExitCodeThread,
	xSwitchToThread,
	xExitThread,
	xGetCurrentThreadId,
	xInitializeCriticalSection,
	xDeleteCriticalSection,
	xEnterCriticalSection,
	xLeaveCriticalSection,
	xInitializeConditionVariable,
	xSleepConditionVariableCS,
	xWakeConditionVariable,
	$

public procedure init()
	kernel32 = open_dll("kernel32.dll")
	xCloseHandle = define_c_func(kernel32, "CloseHandle", {C_HANDLE}, C_BOOL)
	xCreateThread = define_c_func(kernel32, "CreateThread", {C_POINTER,C_SIZE_T,C_POINTER,C_POINTER,C_DWORD,C_POINTER}, C_HANDLE)
	xWaitForSingleObject = define_c_func(kernel32, "WaitForSingleObject", {C_HANDLE,C_DWORD}, C_DWORD)
	xGetExitCodeThread = define_c_func(kernel32, "GetExitCode", {C_HANDLE,C_POINTER}, C_BOOL)
	xSwitchToThread = define_c_func(kernel32, "SwitchToThread", {}, C_BOOL)
	xExitThread = define_c_proc(kernel32, "ExitThread", {C_DWORD})
	xGetCurrentThreadId = define_c_func(kernel32, "GetCurrentThreadId", {}, C_DWORD)
	xInitializeCriticalSection = define_c_proc(kernel32, "InitializeCriticalSection", {C_POINTER})
	xDeleteCriticalSection = define_c_proc(kernel32, "DeleteCriticalSection", {C_POINTER})
	xEnterCriticalSection = define_c_proc(kernel32, "EnterCriticalSection", {C_POINTER})
	xLeaveCriticalSection = define_c_proc(kernel32, "LeaveCriticalSection", {C_POINTER})
	xInitializeConditionVariable = define_c_proc(kernel32, "FuncInitializeConditionVariable", {C_POINTER})
	xSleepConditionVariableCS = define_c_func(kernel32, "SleepConditionVariableCS", {C_POINTER,C_POINTER,C_DWORD}, C_BOOL)
	xWakeConditionVariable = define_c_proc(kernel32, "WakeConditionVariable", {C_POINTER})
end procedure
init()

public function closeHandle(atom Handle)
	return c_func(xCloseHandle, {Handle})
end function

public function createThread(atom pThreadId, atom ThreadProc, atom ptr)
	atom hThread
	hThread = c_func(xCreateThread, {0, 1, ThreadProc, ptr, STACK_SIZE_PARAM_IS_A_RESERVATION, pThreadId})
	if hThread then
		return hThread
	end if
	return -1 -- return negative on error.
end function

public function joinThread(atom hThread, atom lpExitCode)
	atom iret
	iret = c_func(xWaitForSingleObject, {hThread, INFINITE})
	if lpExitCode then
		c_proc(xGetExitCodeThread, {hThread, lpExitCode})
	end if
	return iret
end function

public function yieldThread()
	return c_func(xSwitchToThread, {})
end function

public procedure exitThread(atom dwExitCode)
	c_proc(xExitThread, {dwExitCode})
end procedure

public function selfThread()
	return c_func(xGetCurrentThreadId, {})
end function

public procedure initSection(atom lpCriticalSection)
	c_proc(xInitializeCriticalSection, {lpCriticalSection})
end procedure

public procedure freeSection(atom lpCriticalSection)
	c_proc(xDeleteCriticalSection, {lpCriticalSection})
end procedure

public procedure lockSection(atom pCriticalSection)
	c_proc(xEnterCriticalSection, {pCriticalSection})
end procedure

public procedure unlockSection(atom pCriticalSection)
	c_proc(xLeaveCriticalSection, {pCriticalSection})
end procedure

public procedure initCondition(atom pConditionVariable)
	c_proc(xInitializeConditionVariable, {pConditionVariable})
end procedure

public function sleepCondition(atom pConditionVariable, atom pCriticalSection)
	return c_func(xSleepConditionVariableCS, {pConditionVariable, pCriticalSection, INFINITE})
end function

public function wakeCondition(atom pConditionVariable)
	c_proc(xWakeConditionVariable, {pConditionVariable})
	return 0
end function

public function new_thread(atom cb, atom ptr)
	atom pThreadId, hThread
	pThreadId = allocate_data(4)
	hThread = createThread(pThreadId, cb, ptr)
	return {hThread, pThreadId}
end function

public function new_thread_section()
	atom lpCriticalSection
	lpCriticalSection = allocate_data(SIZE_OF_CRITICAL_SECTION)
	initSection(lpCriticalSection)
	return lpCriticalSection
end function

public function new_thread_condition()
	atom pConditionVariable
	pConditionVariable = allocate_data(SIZE_OF_CONDITION_VARIABLE)
	initCondition(pConditionVariable)
	return pConditionVariable
end function

--end of file.
