// Copyright (C) 2017 James Jonathan Cook
// This program is distributed under the terms of the GNU Lesser General Public License.
//
// This program was written by James Jonathan Cook (jmsck55@gmail.com)
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU Lesser General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
#ifndef euthreadwin_h__
#define euthreadwin_h__

#include <stdio.h>
#include <stdlib.h>

#include <time.h>
#include <unistd.h>

/*
#include <pthread.h>
#include <sched.h>
*/

#include <windows.h>
//#include <synchapi.h>

HMODULE kernel32;

typedef BOOL WINAPI (*FuncCloseHandle)(HANDLE);
typedef HANDLE WINAPI (*FuncCreateThread)(LPSECURITY_ATTRIBUTES,SIZE_T,LPTHREAD_START_ROUTINE,LPVOID,DWORD,LPDWORD);
typedef DWORD WINAPI (*FuncWaitForSingleObject)(HANDLE,DWORD);
typedef BOOL WINAPI (*FuncGetExitCodeThread)(HANDLE,LPDWORD);
typedef BOOL WINAPI (*FuncSwitchToThread)(void);
typedef VOID WINAPI (*FuncExitThread)(DWORD);
typedef DWORD WINAPI (*FuncGetCurrentThreadId)(void);
typedef void WINAPI (*FuncInitializeCriticalSection)(LPCRITICAL_SECTION);
typedef void WINAPI (*FuncDeleteCriticalSection)(LPCRITICAL_SECTION);
typedef void WINAPI (*FuncEnterCriticalSection)(LPCRITICAL_SECTION);
typedef void WINAPI (*FuncLeaveCriticalSection)(LPCRITICAL_SECTION);
typedef void WINAPI (*FuncInitializeConditionVariable)(PCONDITION_VARIABLE);
typedef BOOL WINAPI (*FuncSleepConditionVariableCS)(PCONDITION_VARIABLE,PCRITICAL_SECTION,DWORD);
typedef VOID WINAPI (*FuncWakeConditionVariable)(PCONDITION_VARIABLE);

FuncCloseHandle xCloseHandle;
FuncCreateThread xCreateThread;
FuncWaitForSingleObject xWaitForSingleObject;
FuncGetExitCodeThread xGetExitCodeThread;
FuncSwitchToThread xSwitchToThread;
FuncExitThread xExitThread;
FuncGetCurrentThreadId xGetCurrentThreadId;
FuncInitializeCriticalSection xInitializeCriticalSection;
FuncDeleteCriticalSection xDeleteCriticalSection;
FuncEnterCriticalSection xEnterCriticalSection;
FuncLeaveCriticalSection xLeaveCriticalSection;
FuncInitializeConditionVariable xInitializeConditionVariable;
FuncSleepConditionVariableCS xSleepConditionVariableCS;
FuncWakeConditionVariable xWakeConditionVariable;

// Global variable
typedef DWORD pthread_t;
#define euthread_function(NAME, ARG) DWORD WINAPI NAME( LPVOID ARG )
#define new_mutex(NAME) CRITICAL_SECTION NAME
#define new_condition(NAME) CONDITION_VARIABLE NAME

#ifdef __cplusplus
extern "C" {
#endif

extern void foo(void);

extern void init(void);

//extern

extern int closeHandle(HANDLE h);
extern int createThread(pthread_t * pThreadId, void * ThreadProc, void * arg);
extern int joinThread(int hThread, void * lpExitCode);
extern int yieldThread(void);
extern void exitThread(void * dwExitCode);
extern int selfThread(void);
extern void initSection(LPCRITICAL_SECTION lpCriticalSection);
extern void freeSection(LPCRITICAL_SECTION lpCriticalSection);
extern void lockSection(LPCRITICAL_SECTION lpCriticalSection);
extern void unlockSection(LPCRITICAL_SECTION lpCriticalSection);

extern void initCondition(PCONDITION_VARIABLE pConditionVariable);
extern int sleepCondition(PCONDITION_VARIABLE pConditionVariable, LPCRITICAL_SECTION pCriticalSection);
extern int wakeCondition(PCONDITION_VARIABLE pConditionVariable);


#ifdef __cplusplus
}
#endif

#endif // euthreadwin_h__
