/*
-- Copyright (C) 2017 James Jonathan Cook
-- This program is distributed under the terms of the GNU Lesser General Public License.
--
-- This program was written by James Jonathan Cook (jmsck55@gmail.com)
--
--    This program is free software: you can redistribute it and/or modify
--    it under the terms of the GNU Lesser General Public License as published by
--    the Free Software Foundation, either version 3 of the License, or
--    (at your option) any later version.
--
--    This program is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--    GNU Lesser General Public License for more details.
--
--    You should have received a copy of the GNU Lesser General Public License
--    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
*/

// !!PSUEDO CODE: (PRE-PROCESSOR ANNOTATIONS)

#include <stdio.h>
#include <stdlib.h>

#define mypthread
//#define mywindows

#ifdef mypthread
#include <pthread.h>
#endif

#ifdef mywindows
#include <windows.h>
#endif

#ifdef mypthread
void *functionC();
#endif
#ifdef mywindows
DWORD WINAPI MyThreadFunction( LPVOID lpParameter );
#endif

#ifdef mypthread
pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
#endif
#ifdef mywindows
// Global variable
CRITICAL_SECTION CriticalSection;
#endif
int counter = 0;

int main()
{
#ifdef mypthread
   int rc1, rc2;
   pthread_t thread1, thread2;
	printf("This is mypthread.\n");
#endif
#ifdef mywindows
	DWORD dwThreadId1, dwThreadId2;
	HANDLE hThread1, hThread2;
	printf("This is mywindowsthread.\n");
	// Initialize the critical section one time only.
	if (!InitializeCriticalSectionAndSpinCount(&CriticalSection, 0x00000400) ) {
		printf("Cannot initialize critical section.\n");
	    return 0; }
#endif

printf("Compiled OK.\n");
   /* Create independent threads each of which will execute functionC */

#ifdef mypthread
   if( (rc1=pthread_create( &thread1, NULL, &functionC, NULL)) )
   {
      printf("Thread creation failed: %d\n", rc1);
   }
#endif
#ifdef mywindows
	hThread1 = CreateThread(NULL, 0, MyThreadFunction, NULL, 0, &dwThreadId1);
	if(hThread1 == NULL)
   {
      printf("Thread creation failed: 1\n");
	ExitProcess(2);
   }
#endif

#ifdef mypthread
   if( (rc2=pthread_create( &thread2, NULL, &functionC, NULL)) )
   {
      printf("Thread creation failed: %d\n", rc2);
   }
#endif
#ifdef mywindows
	hThread2 = CreateThread(NULL, 0, MyThreadFunction, NULL, 0, &dwThreadId2);
	if(hThread2 == NULL)
   {
      printf("Thread creation failed: 2\n");
	ExitProcess(2);
   }
#endif

   /* Wait till threads are complete before main continues. Unless we  */
   /* wait we run the risk of executing an exit which will terminate   */
   /* the process and all threads before the threads have completed.   */

#ifdef mypthread
   pthread_join( thread1, NULL);
   pthread_join( thread2, NULL); 
#endif
#ifdef mywindows
//See also: "MsgWaitForMultipleObjectsEx"
	WaitForSingleObject(hThread1, INFINITE);
	WaitForSingleObject(hThread2, INFINITE);
#endif

#ifdef mywindows
	CloseHandle(hThread1);
	CloseHandle(hThread2);
#endif

#ifdef mywindows
	// Release resources used by the critical section object.
	DeleteCriticalSection(&CriticalSection);
#endif
printf("SUCCESS!\n");
   return EXIT_SUCCESS;
}

#ifdef mypthread
void *functionC(void * ptr)
#endif

#ifdef mywindows
DWORD WINAPI MyThreadFunction( LPVOID lpParameter )
#endif

{
#ifdef mypthread
   pthread_mutex_lock( &mutex1 );
#endif
#ifdef mywindows
    // Request ownership of the critical section.
    EnterCriticalSection(&CriticalSection); 
#endif
   counter++;
   printf("Counter value: %d\n",counter);
#ifdef mypthread
   pthread_mutex_unlock( &mutex1 );
#endif
#ifdef mywindows
    // Release ownership of the critical section.
    LeaveCriticalSection(&CriticalSection);
#endif
return 0;
}